;; Generated package description from shell-command+.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "shell-command+" "2.5.0.0.20260516.144445" "An extended shell-command" '((emacs "24.3")) :commit "6d1b35774ef0eda7d0c5f44ca1071ca97d80d0a0" :authors '(("Philip Kaludercic" . "philipk@posteo.net")) :maintainer '("Philip Kaludercic" . "philipk@posteo.net") :keywords '("unix" "processes" "convenience") :url "https://codeberg.org/pkal/shell-command-plus.el")
